from nose.tools import assert_equal
from recipyGui import highlight

def test_highlight():
    """Test the highlight filter"""
    cases = [
        {'txt': 'test', 'q': None, 'out': 'test'},
        {'txt': 'test', 'q': 'test', 'out': '<mark class="no-side-padding">test</mark>'},
        {'txt': None, 'q': None, 'out': None}
    ]

    for c in cases:
        yield assert_equal, highlight(text=c['txt'], query=c['q']), c['out']



import datetime

from tinydb.storages import JSONStorage
from tinydb.serialize import Serializer
from tinydb.middlewares import SerializationMiddleware

class DateTimeSerializer(Serializer):
    OBJ_CLASS = datetime.datetime  # The class this serializer handles

    def encode(self, obj):
        return obj.strftime('%Y-%m-%dT%H:%M:%S')

    def decode(self, s):
        return datetime.strptime(s, '%Y-%m-%dT%H:%M:%S')



serialization = SerializationMiddleware()
serialization.register_serializer(DateTimeSerializer(), 'TinyDate')


import unittest

from recipyCommon import utils

class TestMultipleInsert(unittest.TestCase):

	def test_blank_list(self):
		# Blank list to start with
		l = []
		utils.multiple_insert(l, [1, 2, 3])

		self.assertSequenceEqual(l,
								 [3, 2, 1])

	def test_blank_addition(self):
		# Blank list to start with
		l = [1, 2, 3]
		utils.multiple_insert(l, [])

		self.assertSequenceEqual(l,
								 [1, 2, 3])

	def test_standard_usage(self):
		# Standard usage
		l = ['a', 'b', 'c']
		utils.multiple_insert(l, ['second', 'first'])

		self.assertSequenceEqual(l,
								 ['first', 'second', 'a', 'b', 'c'])




import wrapt
import imp
import os

from tinydb import TinyDB
from .tinydb_serialization import serialization

from .config import get_db_path


def open_or_create_db(path=get_db_path()):
    if not os.path.exists(os.path.dirname(path)):
        os.mkdir(os.path.dirname(path))

    db = TinyDB(path, storage=serialization)

    return db

def multiple_insert(lst, items):
    for item in items:
        lst.insert(0, item)

def recursive_getattr(obj, attr):
    prev_part = obj

    for part in attr.split("."):
        prev_part = getattr(prev_part, part)

    return prev_part

def recursive_setattr(obj, attr, value):
    prev_part = obj

    for part in attr.split(".")[:-1]:
        prev_part = getattr(prev_part, part)

    setattr(prev_part, attr.split(".")[-1], value)

def patch_function(mod, function, wrapper):
        old_f_name = '_%s' % function.replace(".", "_")
        setattr(mod, old_f_name, recursive_getattr(mod, function))

        recursive_setattr(mod, function, wrapper(getattr(mod, old_f_name)))

def create_wrapper(function, arg_loc, source):
    @wrapt.decorator
    def f(self, wrapped, instance, args, kwargs):
        function(args[arg_loc], source)
        return wrapped(*args, **kwargs)

    return f

def recursive_find_module(name, path):
    subnames = name.split(".")
    #print subnames
    #print path
    for subname in subnames[:-1]:
        #print subname
        file_obj, pathname, desc = imp.find_module(subname, path)
        #print subname, file_obj, pathname, desc
        try:
            mod = imp.load_module(subname, file_obj, pathname, desc)
            path = mod.__path__
        finally:
            if file_obj:
                file_obj.close()

    return imp.find_module(subnames[-1], path)




#!/usr/bin/env python
"""recipy - a frictionless provenance tool for Python

Usage:
  recipy search [options] <outputfile>
  recipy latest [options]
  recipy gui [options]
  recipy annotate [options]
  recipy (-h | --help)
  recipy --version

Options:
  -h --help     Show this screen
  --version     Show version
  -a --all      Show all results (otherwise just latest result given)
  -f --fuzzy    Use fuzzy searching on filename
  -r --regex    Use regex searching on filename
  -i --id       Search based on (a fragment of) the run ID
  -v --verbose  Be verbose
  -d --diff     Show diff
  -j --json     Show output as JSON
  --no-browser  Do not open browser window
  --debug       Turn on debugging mode

"""
import os
import re
import sys
import tempfile

from docopt import docopt
from pprint import pprint
from jinja2 import Template
from tinydb import TinyDB, where
from dateutil.parser import parse
from json import dumps
import six

from . import __version__
from recipyCommon import config, utils


db = utils.open_or_create_db()


def template_result(r):
  # Print a single result from the search
    template = """Run ID: {{ unique_id }}
Created by {{ author }} on {{ date }} UTC
Ran {{ script }} using {{ command }}
{% if command_args|length > 0 %}
Using command-line arguments: {{ command_args }}
{% endif %}
{% if gitcommit is defined %}
Git: commit {{ gitcommit }}, in repo {{ gitrepo }}, with origin {{ gitorigin }}
{% endif %}
Environment: {{ environment|join(", ") }}
{% if exception is defined %}
Exception: ({{ exception.type }}) {{ exception.message }}
{% endif %}
{% if inputs|length == 0 %}
Inputs: none
{% else %}
Inputs:
{% for input in inputs %}
  {{ input }}
{% endfor %}
{% endif %}
{% if outputs | length == 0 %}
Outputs: none
{% else %}
Outputs:
{% for output in outputs %}
  {{ output }}
{% endfor %}
{% endif %}

{% if notes is defined %}
Notes:
{{ notes }}
{% endif %}
"""
    template = Template(template, trim_blocks=True)
    return template.render(**r)


def main():
  """
  Main function for recipy command-line script
  """
  args = docopt(__doc__, version='recipy v%s' % __version__)

  if args['--debug']:
      print('Command-line arguments: ')
      print(args)
      print('DB path: ', config.get_db_path())
      print('')
      print('Full config file (as interpreted):')
      print('----------------------------------')
      conf = config.read_config_file()
      s = six.StringIO()
      conf.write(s)
      print(s.getvalue())
      print('----------------------------------')


  if args['search']:
    search(args)
  elif args['latest']:
    latest(args)
  elif args['gui']:
    gui(args)
  elif args['annotate']:
    annotate(args)

def annotate(args):
  # Grab latest run from the DB
  run = get_latest_run()

  # Get temp filename
  f = tempfile.NamedTemporaryFile(delete=False, mode='w')

  # Write something to the bottom of it
  f.write('\n'+'-'*80+'\n')
  f.write('\n')
  f.write('Enter your notes on this run above this line')
  f.write('\n'*3)
  f.write(template_result(run))

  f.close()

  # Open your editor
  os.system('$EDITOR %s' % f.name)

  # Grab the text
  annotation = ""
  with open(f.name, 'r') as f:
    for line in f:
      if line == '-'*80+'\n':
        break
      annotation += line

  notes = annotation.strip()

  if notes == "":
    print('No annotation entered, exiting.')
    return

  # Store in the DB
  db.update({'notes': notes}, where('unique_id') == run['unique_id'])
  db.close()

def gui(args):
  """
  Loads recipy GUI from the command-line
  """
  from recipyGui import recipyGui
  import threading, webbrowser, socket

  def get_free_port():
      port = None
      base_port = config.get_gui_port()
      for trial_port in range(base_port,base_port+5):
          try:
              s = socket.socket()
              s.bind(('', trial_port))
              s.close()
              port = trial_port
              break
          except OSError:
              # port already bound
              pass
      if not port:
          # no free ports above, fall back to random
          s = socket.socket()
          s.bind(('', 0))
          port = s.getsockname()[1]
          s.close()
      return port

  port = get_free_port()
  url = "http://127.0.0.1:{0}".format(port)

  if not args['--no-browser']:
      # Give the application some time before it starts
      threading.Timer(1.25, lambda: webbrowser.open(url) ).start()

  # Turn off reloading by setting debug = False (this also fixes starting the
  # application twice)
  recipyGui.run(debug = args['--debug'], port=port)

def get_latest_run():
  results = db.all()

  results = [_change_date(result) for result in results]

  # Sort the results
  results = sorted(results, key = lambda x: parse(x['date']))

  return results[-1]

def latest(args):

  run = get_latest_run()

  if args['--json']:
    output = dumps(run, indent=2, sort_keys=True)
    print(output)
  else:
    print(template_result(run))

    if args['--diff']:
      if 'diff' in run:
        print("\n\n")
        print(run['diff'])

def search(args):
  filename = args['<outputfile>']

  if args['--fuzzy']:
    results = db.search(where('outputs').any(lambda x: re.match(".+%s.+" % filename, x)))
  elif args['--regex']:
    results = db.search(where('outputs').any(lambda x: re.match(filename, x)))
  elif args['--id']:
    results = db.search(where('unique_id').matches('%s.+' % filename))
    # Automatically turn on display of all results so we don't misleadingly
    # suggest that their shortened ID is unique when it isn't
    args['--all'] = True
  else:
    results = db.search(where('outputs').any(os.path.abspath(filename)))

  results = [_change_date(result) for result in results]

  # Sort the results
  results = sorted(results, key = lambda x: parse(x['date']))

  if args['--json']:
    if args['--all']:
      res_to_output = results
    else:
      res_to_output = results[-1]
    output = dumps(res_to_output, indent=2, sort_keys=True)
    print(output)
  else:
    if len(results) == 0:
        print("No results found")
    else:
        if args['--all']:
            for r in results[:-1]:
                print(template_result(r))
                print("-"*40)
            print(template_result(results[-1]))
        else:
            print(template_result(results[-1]))
            if len(results) > 1:
                print("** Previous runs creating this output have been found. Run with --all to show. **")

            if args['--diff']:
              if 'diff' in results[-1]:
                print("\n\n")
                print(results[-1]['diff'])

  db.close()

def _change_date(result):
  result['date'] = result['date'].replace('{TinyDate}:', '')
  return result

if __name__ == '__main__':
  main()



from flask.ext.wtf import Form
from wtforms import StringField, SubmitField, TextAreaField, HiddenField

class SearchForm(Form):
    query = StringField('query')

class AnnotateRunForm(Form):
    notes = TextAreaField('notes')
    run_id = HiddenField('run_id')



import wrapt
import os
import datetime
import sys
import getpass
import platform
import sys
from traceback import format_tb
from tinydb import TinyDB
import uuid

from git import Repo, InvalidGitRepositoryError

from recipyCommon.config import option_set
from recipyCommon.utils import open_or_create_db

RUN_ID = {}

def get_origin(repo):
    try:
        return repo.remotes.origin.url
    except:
        return None

def new_run():
    log_init()

def log_init():
    # Get the path of the script we're running
    # When running python -m recipy ..., during the recipy import argument 0
    # is -c (for Python 2) or -m (for Python 3) and the script is argument 1
    if sys.argv[0] in ['-c', '-m']:
        # Has the user called python -m recipy without further arguments?
        if len(sys.argv) < 2:
            return
        scriptpath = os.path.realpath(sys.argv[1])
        cmd_args = sys.argv[2:]
    else:
        scriptpath = os.path.realpath(sys.argv[0])
        cmd_args = sys.argv[1:]

    global RUN_ID

    # Open the database
    db = open_or_create_db()

    # Create the unique ID for this run
    guid = str(uuid.uuid4())
    
    
    
    # Get general metadata, environment info, etc
    run = {"unique_id": guid,
        "author": getpass.getuser(),
        "description": "",
        "inputs": [],
        "outputs": [],
        "script": scriptpath,
        "command": sys.executable,
        "environment": [platform.platform(), "python " + sys.version.split('\n')[0]],
        "date": datetime.datetime.utcnow(),
        "command_args": " ".join(cmd_args)}

    if not option_set('ignored metadata', 'git'):
        try:
            repo = Repo(scriptpath, search_parent_directories=True)
            run["gitrepo"] = repo.working_dir
            run["gitcommit"] =  repo.head.commit.hexsha
            run["gitorigin"] = get_origin(repo)

            if not option_set('ignored metadata', 'diff'):
                whole_diff = ''
                diffs = repo.index.diff(None, create_patch=True)
                for diff in diffs:
                    whole_diff += "\n\n\n" + diff.diff.decode("utf-8")

                run['diff'] = whole_diff
        except (InvalidGitRepositoryError, ValueError):
            # We can't store git info for some reason, so just skip it
            pass
    
    # Put basics into DB
    RUN_ID = db.insert(run)

    # Print message
    if not option_set('general', 'quiet'):
        print("recipy run inserted, with ID %s" % (guid))

    db.close()

    # Register exception hook so exceptions can be logged
    sys.excepthook = log_exception

def log_input(filename, source):
    if type(filename) is not str:
        try:
            filename = filename.name
        except:
            pass
    filename = os.path.abspath(filename)
    if option_set('general', 'debug'):
        print("Input from %s using %s" % (filename, source))
    #Update object in DB
    db = open_or_create_db()
    db.update(append("inputs", filename), eids=[RUN_ID])
    db.close()

def log_output(filename, source):
    if type(filename) is not str:
        try:
            filename = filename.name
        except:
            pass
    filename = os.path.abspath(filename)
    if option_set('general', 'debug'):
        print("Output to %s using %s" % (filename, source))
    #Update object in DB
    db = open_or_create_db()
    db.update(append("outputs", filename), eids=[RUN_ID])
    db.close()

def log_update(field, filename, source):
    filename = os.path.abspath(filename)
    print("Adding %s to %s using $s" % (field, filename, source))
    db = open_or_create_db()
    db.update(append(field, filename), eids=[RUN_ID])
    db.close()

def log_exception(typ, value, traceback):
    if option_set('general', 'debug'):
        print("Logging exception %s" % value)
    exception = {'type': typ.__name__,
                 'message': str(value),
                 'traceback': ''.join(format_tb(traceback))}
    # Update object in DB
    db = open_or_create_db()
    db.update({"exception": exception}, eids=[RUN_ID])
    db.close()
    # Done logging, call default exception handler
    sys.__excepthook__(typ, value, traceback)

def append(field, value):
    """
    Append a given value to a given array field.
    Keep an eye on https://github.com/msiemens/tinydb/issues/66
    """
    def transform(element):
        element[field].append(value)

    return transform





from recipyGui import recipyGui
from os import remove
from flask.ext.testing import TestCase
from tinydb import TinyDB
from dateutil.parser import parse

import six

class TestRecipyGui(TestCase):
    def create_app(self):
        self.dbName = 'recipyGui/tests/test.json'
        recipyGui.config['tinydb'] = self.dbName
        recipyGui.config['TESTING'] = True

        self.db = TinyDB(self.dbName)

        self.testRuns = [
            # working run (nothing wrong with it)
            {
                u'inputs': [],
                u'description': u'',
                u'script': u'/Users/janneke/Documents/code/recipy/example_script2.py',
                u'outputs': [u'/Users/janneke/Documents/code/recipy/testNGCM_2.npy'],
                u'author': u'janneke',
                u'gitorigin': u'git@github.com:recipy/recipy.git',
                u'environment': [u'Darwin-13.4.0-x86_64-i386-64bit', u'python 2.7.10 |Anaconda 2.1.0 (x86_64)| (default, May 28 2015, 17:04:42) '],
                u'gitrepo': u'/Users/janneke/Documents/code/recipy',
                u'command': u'/Users/janneke/anaconda/bin/python',
                u'gitcommit': u'a2d8a6a19f33c618943cc7876b5be11fbfe4583b',
                u'date': u'{TinyDate}:2015-08-16T17:20:07',
                u'diff': u"\n\n\n--- a/example_script.py\n+++ b/example_script.py\n@@ -7,5 +7,12 @@ data = pd.read_csv('data.csv')\n data.plot(x='year', y='temperature')\n savefig('newplot.pdf')\n \n+# adding some lines\n+# to test the difff\n+\n data.temperature = data.temperature * 100\n-data.to_csv('output2.csv')\n\\ No newline at end of file\n+data.to_csv('output2.csv')\n+\n+data.to_csv('output3.csv')\n+\n+# more added lines\n\n\n\n--- a/example_script2.py\n+++ b/example_script2.py\n@@ -6,3 +6,5 @@ arr = arr + 500\n # We've made a fairly big change here!\n \n numpy.save('testNGCM_2.npy', arr)\n+\n+# also adding lins to second file\n",
                u'unique_id': u'665a35e7-33eb-4c01-9f78-3560fd31546e'
            },
            # 'empty' run
            {
                u'inputs': [],
                u'description': None,
                u'script': None,
                u'outputs': [],
                u'author': None,
                u'gitorigin': None,
                u'environment': [],
                u'gitrepo': None,
                u'command': None,
                u'gitcommit': None,
                u'date': u'{TinyDate}:2015-08-16T17:20:07',
                u'diff': None,
                u'unique_id': None
            },
            # Run details that caused issue #42
            {
                'author': 'Robin',
                'command': 'C:\\Anaconda3\\python.exe',
                'date': '{TinyDate}:2015-08-22T08:42:59',
                'description': '',
                'diff': '',
                'environment': ['Windows-7-6.1.7601-SP1', 'python 3.4.3 |Anaconda 2.2.0 (64-bit)| (default, Mar  6 2015, 12:06:10) [MSC v.1600 64 bit (AMD64)]'],
                'gitcommit': 'c681315966a9fca0662f57ac45191eb9fdc8cd9c',
                'gitorigin': None,
                'gitrepo': 'C:\\Code\\test-recipy',
                'inputs': [],
                'outputs': ['C:\\Code\\test-recipy\\testNGCM_2.npy'],
                'script': 'C:\\Code\\test-recipy\\example_script2.py',
                'unique_id': '6daab87a-8cf2-4e3d-afb9-1ee5b6b8ecdb'
            },
            # No git info except for the commit (not sure this can happen)
            {
                'author': 'Robin',
                'command': 'C:\\Anaconda3\\python.exe',
                'date': '{TinyDate}:2015-08-22T08:42:59',
                'description': '',
                'diff': None,
                'environment': ['Windows-7-6.1.7601-SP1', 'python 3.4.3 |Anaconda 2.2.0 (64-bit)| (default, Mar  6 2015, 12:06:10) [MSC v.1600 64 bit (AMD64)]'],
                'gitcommit': 'c681315966a9fca0662f57ac45191eb9fdc8cd9c',
                'gitorigin': None,
                'gitrepo': None,
                'inputs': [],
                'outputs': ['C:\\Code\\test-recipy\\testNGCM_2.npy'],
                'script': 'C:\\Code\\test-recipy\\example_script2.py',
                'unique_id': '6daab87a-8cf2-4e3d-afb9-1ee5b6b8ecdb'
            }
        ]

        return recipyGui

    def tearDown(self):
        self.db.close()
        remove(self.dbName)

    def test_index_view_empty_db(self):
        response = self.client.get('/')
        self.assert200(response)

        self.assert_template_used('list.html')
        self.assertContext('runs', [])
        self.assertContext('query', '')

    def test_index_view_db_with_run(self):
        run = self.testRuns[0]
        self.db.insert(run)

        response = self.client.get('/')
        self.assert200(response)

        runs = self.get_context_variable('runs')

        for k, v in six.iteritems(runs[0]):
            self.assertEqual(run.get(k), v)

    def test_index_view_test_runs(self):
        self.testRuns = sorted(self.testRuns, key = lambda x: parse(x['date'].replace('{TinyDate}:', '')) if x['date'] is not None else x['eid'], reverse=True)
        for run in self.testRuns:
            self.db.insert(run)

        response = self.client.get('/')
        self.assert200(response)

        runs = self.get_context_variable('runs')
        self.assertEqual(runs, self.testRuns)

    def test_details_view(self):
        for run in self.testRuns:
            eid = self.db.insert(run)

            response = self.client.get('/run_details?id={}'.format(eid))
            self.assert200(response)

            run2 = self.get_context_variable('run')

            for k, v in six.iteritems(run2):
                self.assertEqual(run.get(k), v)



# Allow any python script to run under recipy by using ::
#
#   python -m recipy SCRIPT [ARGS ...]
from __future__ import print_function

if __name__ == '__main__':
    from sys import argv, exit, stderr
    from runpy import run_path
    # The first argument is the full path to this module so remove it
    argv.pop(0)
    if argv:
        # Run the user script
        run_path(argv[0])
    else:
        print("Usage: python -m recipy SCRIPT [ARGS ...]", file=stderr)
        exit(1)



import os
import unittest
import fake_filesystem_unittest

from recipyCommon import config

class TestExample(fake_filesystem_unittest.TestCase):

    def setUp(self):
        #self.setUpPyfakefs()
        pass

    def tearDown(self):
        # It is no longer necessary to add self.tearDownPyfakefs()
        pass

    def test_create_file(self):
        # # The os module has been replaced with the fake os module so all of the
        # # following occurs in the fake filesystem.
        # self.assertFalse(os.path.isdir('/test'))
        # os.mkdir('/test')
        # self.assertTrue(os.path.isdir('/test'))

        # self.assertFalse(os.path.exists('/test/file.txt'))
        # example.create_file('/test/file.txt')
        # self.assertTrue(os.path.exists('/test/file.txt'))
        pass



import sys
from .PatchImporter import PatchImporter
from .PatchSimple import PatchSimple
import wrapt

from .log import log_input, log_output
from recipyCommon.utils import create_wrapper, multiple_insert

class PatchPandas(PatchSimple):
    modulename = 'pandas'
    input_functions = ['read_csv', 'read_table', 'read_excel', 'read_hdf', 'read_pickle',
        'read_stata', 'read_msgpack']

    output_functions = ['DataFrame.to_csv', 'DataFrame.to_excel', 'DataFrame.to_hdf',
        'DataFrame.to_msgpack', 'DataFrame.to_stata', 'DataFrame.to_pickle']

    output_functions += ['Panel.to_excel', 'Panel.to_hdf',
        'Panel.to_msgpack', 'Panel.to_pickle']

    output_functions += ['Series.to_csv', 'Series.to_hdf',
        'Series.to_msgpack', 'Series.to_pickle']

    input_wrapper = create_wrapper(log_input, 0, 'pandas')
    output_wrapper = create_wrapper(log_output, 0, 'pandas')

class PatchMPL(PatchSimple):
    modulename = 'matplotlib.pyplot'

    input_functions = []
    output_functions = ['savefig']

    input_wrapper = create_wrapper(log_input, 0, 'matplotlib')
    output_wrapper = create_wrapper(log_output, 0, 'matplotlib')

class PatchNumpy(PatchSimple):
    modulename = 'numpy'

    # The `load` function is *deliberately* not included here
    # as it calls fromfile internally, and then we get two duplicate
    # entries recorded in the log
    input_functions = ['genfromtxt', 'loadtxt', 'fromfile']
    output_functions = ['save', 'savez', 'savez_compressed', 'savetxt']

    input_wrapper = create_wrapper(log_input, 0, 'numpy')
    output_wrapper = create_wrapper(log_output, 0, 'numpy')

multiple_insert(sys.meta_path, [PatchPandas(), PatchMPL(), PatchNumpy()])


import sys
import imp

import functools
from recipyCommon.config import option_set
from recipyCommon.utils import recursive_find_module

class PatchImporter(object):
    """A class that handles finding modules, importing them
    and calling a `patch` method.

    This class is not designed to be used itself - instead,
    subclasses should be created that implement the `patch` method.
    """
    modulename = ''

    def find_module(self, fullname, path=None):
        """Module finding method. It tells Python to use our hook
        only for the package we want.
        """
        #print("Find module: %s" % fullname)
        if fullname == self.modulename:
            self.path = path
            return self
        return None
    
    def load_module(self, name):
        """Module loading method. It imports the module normally,
        and then calls the `patch` method to wrap the functions we need.

        `patch` is implemented by subclasses
        """
        if name != self.modulename:
            raise ImportError("%s can only be used to import a specific module!",
                              self.__class__.__name__)
        if name in sys.modules:
            return sys.modules[name]    # already imported and patched
        
        # Find the module
        file_obj, pathname, desc = recursive_find_module(name, sys.path)

        try:
            mod = imp.load_module(name, file_obj, pathname, desc)
        finally:
            if file_obj:
                file_obj.close()
        
        if option_set('general', 'debug'):
            print("Patching %s" % mod.__name__)

        # Actually do the patching
        mod = self.patch(mod)

        # And put the module in Python's proper namespace
        sys.modules[name] = mod

        return mod

    
    def patch(self, mod):
        return mod



import sys
from .PatchImporter import PatchImporter
from .PatchSimple import PatchSimple

from .log import log_input, log_output
from recipyCommon.utils import create_wrapper, multiple_insert

class PatchGDAL(PatchSimple):
    modulename = 'gdal'

    input_functions = ['Open']
    output_functions = ['Driver.Create', 'Driver.CreateCopy']

    input_wrapper = create_wrapper(log_input, 0, 'gdal')
    output_wrapper = create_wrapper(log_output, 0, 'gdal')

class PatchSKLearn(PatchSimple):
    modulename = 'sklearn'

    input_functions = ['datasets.load_svmlight_file']
    output_functions = ['datasets.dump_svmlight_file']

    input_wrapper = create_wrapper(log_input, 0, 'sklearn')
    output_wrapper = create_wrapper(log_output, 0, 'sklearn')

class PatchSKImage(PatchSimple):
    modulename = 'skimage'

    input_functions = ['io.imread', 'io.load_sift', 'io.load_surf', 'external.tifffile.imread']
    output_functions = ['io.imsave', 'external.tifffile.imsave']

    input_wrapper = create_wrapper(log_input, 0, 'skimage')
    output_wrapper = create_wrapper(log_output, 0, 'skimage')

# class PatchPillow(PatchSimple):
#     modulename = 'PIL'

#     input_functions = ['Image.open']
#     output_functions = ['Image.save']

#     input_wrapper = create_wrapper(log_input, 0, 'Pillow')
#     output_wrapper = create_wrapper(log_output, 0, 'Pillow')

class PatchNIBabel(PatchSimple):
    modulename = 'nibabel'

    images = ['nifti1.Nifti1Image', 'nifti2.Nifti2Image', 'freesurfer.mghformat.MGHImage', 'spm99analyze.Spm99AnalyzeImage', 'minc1.Minc1Image', 'minc2.Minc2Image', 'analyze.AnalyzeImage', 'parrec.PARRECImage', 'spm2analyze.Spm2AnalyzeImage']
    input_functions = [image_name + '.from_filename' for image_name in images]
    output_functions = [image_name + '.to_filename' for image_name in images]

    input_wrapper = create_wrapper(log_input, 0, 'nibabel')
    output_wrapper = create_wrapper(log_output, 0, 'nibabel')

multiple_insert(sys.meta_path, [PatchGDAL(), PatchSKLearn(), PatchSKImage(),
                                PatchNIBabel()])



import sys
from .PatchImporter import PatchImporter
import wrapt

from .log import *
from recipyCommon.utils import *
from recipyCommon.config import option_set

class PatchSimple(PatchImporter):
    """Simple sublass of PatchImporter which implements
    a simple `patch` method that just patches specific input
    and output functions, as specified in the attributes of the class.

    This should not be used directly, but subclasses which set these attributes
    should be created.
    """

    def patch(self, mod):
        """Do the patching of `input_functions` and `output_functions`
        in `mod` using `input_wrapper` and `output_wrapper` respectively.
        """
        
        if not self._ignore_input():
            for f in self.input_functions:
                if option_set('general', 'debug'):
                    print('Patching input function: %s' % f)
                patch_function(mod, f, self.input_wrapper)
        else:
            if option_set('general', 'debug'):
                    print('Ignoring inputs for: %s' % self.modulename)

        if not self._ignore_output():
            for f in self.output_functions:
                if option_set('general', 'debug'):
                    print('Patching output function: %s' % f)
                patch_function(mod, f, self.output_wrapper)
        else:
            if option_set('general', 'debug'):
                    print('Ignoring outputs for: %s' % self.modulename)

        return mod

    def _ignore_input(self):
        root_modulename = self.modulename.split('.')[0]

        return option_set('ignored inputs', root_modulename) or option_set('ignored inputs', 'all')

    def _ignore_output(self):
        root_modulename = self.modulename.split('.')[0]

        return option_set('ignored outputs', root_modulename) or option_set('ignored outputs', 'all')


from flask import Flask, url_for
import os
from flask_bootstrap import Bootstrap
import re
from time import strptime, strftime

from recipyCommon.config import get_db_path

recipyGui = Flask(__name__)
recipyGui.config['SECRET_KEY'] = 'geheim'
recipyGui.config['tinydb'] = get_db_path()

Bootstrap(recipyGui)

# Function to easily find your assets
# In your template use <link rel=stylesheet href="{{ static('filename') }}">
recipyGui.jinja_env.globals['static'] = (
    lambda filename: url_for('static', filename = filename)
)

def register_blueprints(app):
    # Prevents circular imports
    from recipyGui.views import routes
    recipyGui.register_blueprint(routes)

register_blueprints(recipyGui)

# Custom filters
@recipyGui.template_filter()
def highlight(text, query=None):
    """Filter to highlight query terms in search results."""
    if query:
        replacement = r'<mark class="no-side-padding">\1</mark>'
        for q in query.split(' '):
            text = re.sub(r'(?i)({})'.format(q), replacement, text)
    return text

recipyGui.jinja_env.filters['highlight'] = highlight

@recipyGui.template_filter()
def datetimefilter(value, format='%Y/%m/%d %H:%M'):
    """convert a datetime to a different format."""
    value = strptime(value, '{TinyDate}:%Y-%m-%dT%H:%M:%S')
    return strftime(format, value) + " UTC"

recipyGui.jinja_env.filters['datetimefilter'] = datetimefilter

@recipyGui.template_filter()
def gitorigin2url(origin):
    """convert a datetime to a different format."""
    url = origin.replace(':', '/')
    url = url.replace('git@', 'http://')
    url = url.replace('.git', '')

    return url

recipyGui.jinja_env.filters['gitorigin2url'] = gitorigin2url

@recipyGui.template_filter()
def colordiff(diff):
    """convert git diff data to html/bootstrap color code"""
    if diff=='':
        return ''
    diff = diff.strip()
    diff = diff.replace('\n', '&nbsp;\n')
    diffData = diff.split('\n')
    print(diffData)
    openTag = '<tr><td class="'
    openTagEnd = '">'
    nbsp = '&nbsp;&nbsp;&nbsp;&nbsp;'
    data = '\n'.join([('%s%s%s%s<samp>%s</samp><td></tr>' % (openTag, 'bg-danger' if line.startswith('-') else ('bg-success' if line.startswith('+') else ('bg-info' if line.startswith('@') else '')), openTagEnd, nbsp*line.count('\t') ,line)) for line in diffData])
    return '<table width="100%">'+data+'</table>'

recipyGui.jinja_env.filters['colordiff'] = colordiff



from flask import Blueprint, request, render_template, redirect, url_for, \
    escape, make_response
from recipyGui import recipyGui
from .forms import SearchForm, AnnotateRunForm
from tinydb import TinyDB, where
import re
from dateutil.parser import parse
import os
from ast import literal_eval
from json import dumps


routes = Blueprint('routes', __name__, template_folder='templates')

if not os.path.exists(os.path.dirname(recipyGui.config.get('tinydb'))):
        os.mkdir(os.path.dirname(recipyGui.config.get('tinydb')))


@recipyGui.route('/')
def index():
    form = SearchForm()

    query = request.args.get('query', '').strip()

    db = TinyDB(recipyGui.config.get('tinydb'))

    if not query:
        runs = db.all()
    else:
        # Search run outputs using the query string
        runs = db.search(
            where('outputs').any(lambda x: re.match(".+%s.+" % query, x)) |
            where('notes').contains(query) |
            where('unique_id').contains(query))
    runs = sorted(runs, key = lambda x: parse(x['date'].replace('{TinyDate}:', '')) if x['date'] is not None else x['eid'], reverse=True)

    run_ids = []
    for run in runs:
        if 'notes' in run.keys():
            run['notes'] = str(escape(run['notes']))
        run_ids.append(run.eid)

    db.close()

    return render_template('list.html', runs=runs, query=query, form=form,
                           run_ids=str(run_ids))


@recipyGui.route('/run_details')
def run_details():
        form = SearchForm()
        annotateRunForm = AnnotateRunForm()
        query = request.args.get('query', '')
        run_id = int(request.args.get('id'))

        db = TinyDB(recipyGui.config.get('tinydb'))
        r = db.get(eid=run_id)

        db.close()

        return render_template('details.html', query=query, form=form,
                               annotateRunForm=annotateRunForm, run=r)


@recipyGui.route('/latest_run')
def latest_run():
    form = SearchForm()
    annotateRunForm = AnnotateRunForm()

    db = TinyDB(recipyGui.config.get('tinydb'))

    runs = db.all()
    runs = sorted(runs, key = lambda x: parse(x['date'].replace('{TinyDate}:', '')), reverse=True)

    db.close()

    return render_template('details.html', query='', form=form, run=runs[0],
                           annotateRunForm=annotateRunForm,
                           active_page='latest_run')

@recipyGui.route('/annotate', methods=['POST'])
def annotate():
    notes = request.form['notes']
    run_id = int(request.form['run_id'])

    query = request.args.get('query', '')

    db = TinyDB(recipyGui.config.get('tinydb'))
    db.update({'notes': notes}, eids=[run_id])
    db.close()

    return redirect(url_for('run_details', id=run_id, query=query))

@recipyGui.route('/runs2json', methods=['POST'])
def runs2json():
    run_ids = literal_eval(request.form['run_ids'])
    db = TinyDB(recipyGui.config.get('tinydb'))
    runs = [db.get(eid=run_id) for run_id in run_ids]
    db.close()

    response = make_response(dumps(runs, indent=2, sort_keys=True))
    response.headers['content-type'] = 'application/json'
    response.headers['Content-Disposition'] = 'attachment; filename=runs.json'
    return response



# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

setup(
    name='recipy',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version='0.2.1',

    description='A frictionless provenance framework for Python',
    long_description="""A frictionless provenance framework for Python.

Please see https://github.com/recipy/recipy for further information.
""",

    # The project's main homepage.
    url='https://github.com/recipy/recipy',

    # Author details
    author='Robin Wilson, Raquel Alegre, Janneke van der Zwaan',
    author_email='robin@rtwilson.com',

    # Choose your license
    license='Apache',

    include_package_data=True,

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: Apache Software License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
    ],

    # What does your project relate to?
    keywords='development, science, reproducibility, provenance',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(),

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=['wrapt', 'tinydb', 'jinja2', 'docopt', 'GitPython', 'Flask',
                      'Flask-Script', 'flask_bootstrap', 'flask-wtf', 'python-dateutil',
                      'six'],

    entry_points={
        'console_scripts': [
            'recipy=recipyCmd.recipycmd:main',
        ]
    }

    #scripts=['recipy-cmd']
)


"""
Configuration file-related functions for recipy
"""

try:
    from ConfigParser import SafeConfigParser, Error
except:
    from configparser import SafeConfigParser, Error

import os

def read_config_file():
    """
    Reads the recipy configuration file, which is in Windows INI-style format

    Try .recipyrc and recipy in the current directory, and then ~/.recipy/recipyrc
    """
    CONFIG = SafeConfigParser(allow_no_value=True)
    
    # Try reading files in the current directory first
    files_read = CONFIG.read(['.recipyrc', 'recipyrc'])

    # Read the config file in the home directory
    if len(files_read) == 0:
        CONFIG.read(os.path.expanduser("~/.recipy/recipyrc"))

    return CONFIG


conf = read_config_file()

def option_set(section, name):
    return conf.has_option(section, name)

def get_db_path():
    try:
        return conf.get('database', 'path')
    except Error:
        return os.path.expanduser('~/.recipy/recipyDB.json')
    
def get_gui_port():
    try:
        return int(conf.get('general', 'port'))
    except Error:
        return 9000
